package main

import (
	"cs"
	"errors"
	"fmt"
	"math/rand"
)

type GuildMgr struct {
	cs.MsgHandlerImpl
	Guilds            map[int]*Guild      // 公会ID -> 公会列表
	PendingAccountIds map[int][]string    // 公会ID -> 申请者账户ID列表
	Logs              map[int][]*GuildLog // 公会ID -> 公会日志
}

func NewGuildMgr() *GuildMgr {
	return &GuildMgr{
		Guilds:            make(map[int]*Guild),
		PendingAccountIds: make(map[int][]string),
		Logs:              make(map[int][]*GuildLog),
	}
}

var guildID = 0

// CreateGuild 创建公会
func (gm *GuildMgr) CreateGuild(name string, leaderID string, maxMembers int) (int, error) {
	if maxMembers <= 0 {
		return 0, errors.New("invalid maxMembers")
	}

	if _, exists := gm.Guilds[guildID]; exists {
		return 0, errors.New("guild already exists")
	}
	guildID++

	newGuild := &Guild{
		ID:         guildID,
		Name:       name,
		LeaderID:   leaderID,
		MaxMembers: maxMembers,
	}

	gm.Guilds[guildID] = newGuild
	return guildID, nil
}

// ApplyToGuild 申请加入公会
func (gm *GuildMgr) ApplyToGuild(guildID int, applicantAccountId string) error {
	if _, exists := gm.Guilds[guildID]; !exists {
		return errors.New("guild not found")
	}

	guild := gm.Guilds[guildID]
	if len(guild.Members) >= guild.MaxMembers {
		return errors.New("guild is full")
	}

	gm.PendingAccountIds[guildID] = append(gm.PendingAccountIds[guildID], applicantAccountId)
	return nil
}

// ApproveApplication 批准申请加入公会
func (gm *GuildMgr) ApproveApplication(guildID int, applicantAccountId string) error {
	guild, exists := gm.Guilds[guildID]
	if !exists {
		return errors.New("guild not found")
	}

	for guildId, guild := range gm.Guilds {
		for _, member := range guild.Members {
			if applicantAccountId == member.AccountId {
				errMsg := fmt.Sprintf("Player:%s has joined other guild:%d", applicantAccountId, guildId)
				return errors.New(errMsg)
			}
		}
	}

	pendingAccountIds, pendingExists := gm.PendingAccountIds[guildID]
	if !pendingExists {
		return errors.New("no pending applications for this guild")
	}

	for i, accountId := range pendingAccountIds {
		if accountId == applicantAccountId {
			newMember := GuildMember{AccountId: applicantAccountId, Level: 3}
			err := guild.AddMember(newMember)
			if err != nil {
				return err
			}
			gm.PendingAccountIds[guildID] = append(pendingAccountIds[:i], pendingAccountIds[i+1:]...)
			return nil
		}
	}

	return errors.New("applicant not found")
}

// RejectApplication 拒绝申请加入公会
func (gm *GuildMgr) RejectApplication(guildID int, applicantAccountId string) error {
	pendingAccountIds, exists := gm.PendingAccountIds[guildID]
	if !exists {
		return errors.New("no pending applications for this guild")
	}

	for i, accountId := range pendingAccountIds {
		if accountId == applicantAccountId {
			gm.PendingAccountIds[guildID] = append(pendingAccountIds[:i], pendingAccountIds[i+1:]...)
			return nil
		}
	}

	return errors.New("applicant not found")
}

// JoinGuild 可直接加入公会
func (gm *GuildMgr) JoinGuild(guildID int, memberID string) error {
	guild, exists := gm.Guilds[guildID]
	if !exists {
		return errors.New("guild not found")
	}

	if len(guild.Members) >= guild.MaxMembers {
		return errors.New("guild is full")
	}

	newMember := GuildMember{AccountId: memberID, Level: 3}
	guild.Members = append(guild.Members, newMember)
	return nil
}

// LeaveGuild 离开公会
func (gm *GuildMgr) LeaveGuild(guildID int, memberID string) error {
	guild, exists := gm.Guilds[guildID]
	if !exists {
		return errors.New("guild not found")
	}
	if guild.LeaderID == memberID {
		return errors.New("leader not leave guild")
	}

	for i, member := range guild.Members {
		if member.AccountId == memberID {
			guild.Members = append(guild.Members[:i], guild.Members[i+1:]...)
			return nil
		}
	}

	return errors.New("member not found in the guild")
}

// DismissMember 开除成员 踢出
func (gm *GuildMgr) DismissMember(guildID int, operatorAccountId, memberID string) error {
	guild, exists := gm.Guilds[guildID]
	if !exists {
		return errors.New("guild not found")
	}

	if memberID == guild.LeaderID {
		return errors.New("cannot dismiss the leader")
	}

	operatorMember, err := guild.GetMember(operatorAccountId)
	if err != nil {
		return errors.New("cannot dismiss the leader")
	}
	if operatorMember.Level >= 2 {
		return errors.New("no permission")
	}

	dismissMember, err := guild.GetMember(memberID)
	if err != nil {
		return errors.New("member not found in the guild")
	}
	if dismissMember.Level <= 2 {
		return errors.New("no permission")
	}

	_ = guild.RemoveMember(memberID)

	return nil
}

// PromoteMember 提升成员为干部
func (gm *GuildMgr) PromoteMember(guildID int, memberID string) error {
	guild, exists := gm.Guilds[guildID]
	if !exists {
		return errors.New("guild not found")
	}

	if memberID == guild.LeaderID {
		return errors.New("cannot promote the leader")
	}

	for _, member := range guild.Members {
		if member.AccountId == memberID {
			member.Level = 2
			return nil
		}
	}

	return errors.New("member not found in the guild")
}

// DemoteMember 解除成员干部身份
func (gm *GuildMgr) DemoteMember(guildID int, memberID string) error {
	guild, exists := gm.Guilds[guildID]
	if !exists {
		return errors.New("guild not found")
	}

	if memberID == guild.LeaderID {
		return errors.New("cannot demote the leader")
	}

	for _, member := range guild.Members {
		if member.AccountId == memberID {
			member.Level = 3
			return nil
		}
	}

	return errors.New("member not found in the guild")
}

// WriteLog 记录公会日志
func (gm *GuildMgr) WriteLog(guildID int, accountId string, logType uint16, content string) {
	_, exists := gm.Logs[guildID]
	if !exists {
		gm.Logs[guildID] = make([]*GuildLog, 0)
	}
	log := &GuildLog{
		GuildId:   guildID,
		AccountId: accountId,
		LogType:   logType,
		Content:   content,
	}
	gm.Logs[guildID] = append(gm.Logs[guildID], log)
}

// Disband 解散公会
func (gm *GuildMgr) Disband(guildID int, accountId string) error {
	guild, exists := gm.Guilds[guildID]
	if !exists {
		return errors.New("guild not found")
	}

	if accountId != guild.LeaderID {
		return errors.New("cannot disband guild")
	}

	guildName := guild.Name
	guild.Members = nil

	gm.Guilds[guildID] = nil
	gm.PendingAccountIds[guildID] = nil
	gm.Logs[guildID] = nil

	// 确保在删除之前没有其他地方引用了公会对象的指针, 以避免空指针异常
	delete(gm.Guilds, guildID)
	delete(gm.PendingAccountIds, guildID)
	delete(gm.Logs, guildID)

	gm.WriteLog(guildID, accountId, 1, "Guild disbanded: "+guildName)

	return nil
}

// SearchGuilds 根据关键字搜索公会
func (gm *GuildMgr) SearchGuilds(keyword string) []Guild {
	var results []Guild
	for _, guild := range gm.Guilds {
		if containsSubstring(guild.Name, keyword) {
			results = append(results, *guild)
		}
	}
	return results
}

// RandomGuilds 随机10个公会
func (gm *GuildMgr) RandomGuilds() []Guild {
	randomGuildNum := 10
	count := len(gm.Guilds)
	if count <= randomGuildNum {
		randomGuildNum = count
	}
	// shuffle 从count50，返回一个slice []int, 包含10个元素
	randIndices := rand.Perm(count)[:randomGuildNum]
	var results []Guild
	for _, idx := range randIndices {
		results = append(results, *gm.Guilds[idx])
	}
	return results
}

// GetGuildByAccountId 查询我的工会
func (gm *GuildMgr) GetGuildByAccountId(accountId string) *Guild {
	for _, guild := range gm.Guilds {
		for _, member := range guild.Members {
			if accountId == member.AccountId {
				return guild
			}
		}
	}
	return nil
}

// Info 工会信息
func (gm *GuildMgr) Info(accountId string) (*Guild, []string, []*GuildLog) {
	guild := gm.GetGuildByAccountId(accountId)
	pendingAccountIds := make([]string, 10)
	guildLogs := make([]*GuildLog, 10)
	if guild != nil {
		guildId := guild.ID
		pendingAccountIds = gm.PendingAccountIds[guildId]
		guildLogs = gm.Logs[guildId]
	}

	return guild, pendingAccountIds, guildLogs
}

func containsSubstring(s, substr string) bool {
	return len(s) >= len(substr) && s[len(s)-len(substr):] == substr
}
